/******************************************************************************
 * $Id: gdal_gimp.c $
 *
 * GDAL plugin for GIMP
 *
 ******************************************************************************
 * Copyright (c) 2008, Even Rouault
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

/* How to compile and install ?

Compile with :
gcc -Wall gimp-gdal.c -o gimp-gdal `gdal-config --cflags` `gdal-config --libs` `pkg-config --cflags --libs gimp-2.0 gimpui-2.0 gtk+-2.0`

Install by copying gimp-gdal into /usr/lib/gimp/2.0/plug-ins

You can force the opening of a dataset through this plugin by adding 'GDAL:' in front of the filename.
For example gimp GDAL:foo.tif will force foo.tif to be opened by GDAL and not the regular TIFF plugin.

*/


#include <stdlib.h>
#include <string.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include <gdal.h>
#include <cpl_string.h>
#include <cpl_error.h>

#define LOAD_PROC "file-gdal-load"

static gint32             ReadGDAL   (const gchar *filename);

gboolean     interactive = FALSE;

/* Declare some local functions.
 */
static void   query (void);
static void   run   (const gchar      *name,
                     gint              nparams,
                     const GimpParam  *param,
                     gint             *nreturn_vals,
                     GimpParam       **return_vals);

const GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

MAIN ()

static void
query (void)
{
  static const GimpParamDef load_args[] =
  {
    { GIMP_PDB_INT32,    "run-mode",     "Interactive, non-interactive" },
    { GIMP_PDB_STRING,   "filename",     "The name of the file to load" },
    { GIMP_PDB_STRING,   "raw-filename", "The name entered" },
  };
  static const GimpParamDef load_return_vals[] =
  {
    { GIMP_PDB_IMAGE, "image", "Output image" },
  };

  gimp_install_procedure (LOAD_PROC,
                          "Loads GDAL files",
                          "Loads GDAL files",
                          "Even Rouault",
                          "Even Rouault",
                          "2008",
                          "GDAL dataset",
                          NULL,
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (load_args),
                          G_N_ELEMENTS (load_return_vals),
                          load_args, load_return_vals);

  gimp_register_file_handler_mime (LOAD_PROC, "image/gdal");
  /* Add more extension here ! */
  gimp_register_magic_load_handler (LOAD_PROC,
                                    "dt0,dt1,dt2,tif,tiff,img,kap,vrt,ovr,gen,ntf,jp2,h5,toc,*",
                                    "GDAL:,NITF_TOC_ENTRY:,HDF5:,GTIFF_DIR:,GTIFF_RAW:",
                                    "0,string,NITF");

}

static void
run (const gchar      *name,
     gint             nparams,
     const GimpParam  *param,
     gint             *nreturn_vals,
     GimpParam       **return_vals)
{
  static GimpParam  values[2];
  GimpRunMode       run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;
  gint32            image_ID;

  run_mode = param[0].data.d_int32;

  /* INIT_I18N (); */

  *nreturn_vals = 1;
  *return_vals  = values;
  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;

  if (strcmp (name, LOAD_PROC) == 0)
    {
       switch (run_mode)
        {
        case GIMP_RUN_INTERACTIVE:
          interactive = TRUE;
          break;

        case GIMP_RUN_NONINTERACTIVE:
          /*  Make sure all the arguments are there!  */
          if (nparams != 3)
            status = GIMP_PDB_CALLING_ERROR;
          break;

        default:
          break;
        }

       if (status == GIMP_PDB_SUCCESS)
         {
           image_ID = ReadGDAL (param[1].data.d_string);

           if (image_ID != -1)
             {
               *nreturn_vals = 2;
               values[1].type         = GIMP_PDB_IMAGE;
               values[1].data.d_image = image_ID;
             }
           else
             {
               status = GIMP_PDB_EXECUTION_ERROR;
             }
         }
    }
  else
    {
      status = GIMP_PDB_CALLING_ERROR;
    }

  values[0].data.d_status = status;
}

static gint32 ReadGDAL (const gchar *filename)
{
  GDALDatasetH       hDS;
  int                nBands;
  int                width, height;
  GDALColorTableH    hColorTable;
  GimpPixelRgn       pixel_rgn;
  gint32             image;
  gint32             layer;
  GimpDrawable      *drawable;
  guchar            *dest;
  guchar             gimp_cmap[768];
  glong              channels;
  gint               i;
  GimpImageBaseType  base_type;
  GimpImageType      image_type;

  GDALAllRegister();

  if (strncmp(filename, "GDAL:", 5) == 0)
      filename += 5;

  hDS = GDALOpen(filename, GA_ReadOnly);

  if (hDS == NULL)
    {
      g_message ("Could not open '%s' for reading: %s",
                 gimp_filename_to_utf8 (filename), CPLGetLastErrorMsg());
      return -1;
    }

  nBands = GDALGetRasterCount(hDS);
  if (nBands == 0)
  {
      char** papszSubdatasets = GDALGetMetadata(hDS, "SUBDATASETS");
      if (papszSubdatasets != NULL && interactive)
      {
        GtkWidget *dialog;
        GtkWidget *main_vbox;
        GtkWidget *alignment;
        GtkWidget *main_hbox;
        GtkWidget *combo_box;
        gboolean   run;

        gimp_ui_init ("gimp-gdal-select-subdataset", FALSE);

        dialog = gimp_dialog_new ("Select a subdataset", "gimp-gdal-select-subdataset",
                                NULL, 0,
                                NULL, NULL,
                                GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                GTK_STOCK_OK,     GTK_RESPONSE_OK,
                                NULL);

        main_vbox = gtk_vbox_new (FALSE, 6);
        gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
        gtk_widget_show (main_vbox);

        alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
        gtk_widget_show (alignment);
        gtk_container_add (GTK_CONTAINER (main_vbox), alignment);
        gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 6, 6, 6, 6);

        main_hbox = gtk_hbox_new (FALSE, 0);
        gtk_widget_show (main_hbox);
        gtk_container_add (GTK_CONTAINER (alignment), main_hbox);

        combo_box = gtk_combo_box_new_text();
        gtk_widget_show (combo_box);
        gtk_container_add (GTK_CONTAINER (main_hbox), combo_box);

        char** iter = papszSubdatasets;
        while(*iter)
        {
            gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box), strchr(iter[1], '=') + 1);
            iter += 2;
        }
        gtk_combo_box_set_active(GTK_COMBO_BOX(combo_box), 0);

        gtk_widget_show (dialog);

        run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);

        if (run)
        {
            int iActive = gtk_combo_box_get_active(GTK_COMBO_BOX(combo_box));
            filename = CPLStrdup(strchr(papszSubdatasets[2 * iActive], '=') + 1);
            GDALClose(hDS);
            hDS = GDALOpen(filename, GA_ReadOnly);
            nBands = GDALGetRasterCount(hDS);
        }

        gtk_widget_destroy (dialog);

        if (!run)
        {
            return -1;
        }
      }
      else
      {
        g_message ("Dataset with invalid number of band : %d.", nBands);
        return -1;
      }
  }

  gimp_progress_init_printf ("Opening '%s'",
                             gimp_filename_to_utf8 (filename));

  width = GDALGetRasterXSize(hDS);
  height = GDALGetRasterYSize(hDS);
  hColorTable = GDALGetRasterColorTable(GDALGetRasterBand(hDS, 1));

  if (nBands == 1)
  {
      if (hColorTable == NULL)
      {
          base_type = GIMP_GRAY;
          image_type = GIMP_GRAY_IMAGE;
      }
      else
      {
          base_type = GIMP_INDEXED;
          image_type = GIMP_INDEXED_IMAGE;
      }
      channels = 1;
  }
  else if (nBands == 3)
  {
      base_type = GIMP_RGB;
      image_type = GIMP_RGB_IMAGE;
      channels = 3;
  }
  else if (nBands == 4)
  {
      base_type = GIMP_RGB;
      image_type = GIMP_RGBA_IMAGE;
      channels = 4;
  }
  else
  {
      g_message ("Dataset with invalid number of band : %d.", nBands);
      return -1;
  }

  image = gimp_image_new (width, height, base_type);
  layer = gimp_layer_new (image, "Background",
                          width, height,
                          image_type, 100, GIMP_NORMAL_MODE);

  gimp_image_set_filename (image, filename);

  gimp_image_add_layer (image, layer, 0);
  drawable = gimp_drawable_get (layer);

  dest      = g_malloc (drawable->width * channels);

  for(i = 0; i < height; i ++)
  {
      GDALDatasetRasterIO(hDS, GF_Read, 0, i, width, 1,
                          dest, width, 1,
                          GDT_Byte, channels, NULL, channels,  drawable->width * channels, 1);
      gimp_pixel_rgn_init (&pixel_rgn, drawable,
                       0, i, drawable->width, 1, TRUE, FALSE);
      gimp_pixel_rgn_set_rect (&pixel_rgn, dest, 0, i, drawable->width, 1);
      gimp_drawable_flush (drawable);
      gimp_progress_update ((gdouble) i /
                                    (gdouble) height);
  }

  if (hColorTable != NULL)
  {
      for (i = 0; i < GDALGetColorEntryCount(hColorTable); i++)
      {
          GDALColorEntry colorEntry;
          GDALGetColorEntryAsRGB(hColorTable, i, &colorEntry);
          gimp_cmap[3 * i + 0] = colorEntry.c1;
          gimp_cmap[3 * i + 1] = colorEntry.c2;
          gimp_cmap[3 * i + 2] = colorEntry.c3;
      }
  }

  gimp_progress_update (1);

  if (hColorTable != NULL)
    gimp_image_set_colormap (image, gimp_cmap, GDALGetColorEntryCount(hColorTable));

  gimp_drawable_detach (drawable);
  g_free (dest);

  GDALClose(hDS);

  return image;
}
